﻿using System.ComponentModel;
using Infragistics.Win;
using Infragistics.Win.UltraMessageBox;
using Infragistics.Win.UltraWinGrid;

namespace PopupControlContainerPlayground
{

	#region Public Delegates

	/// <summary>
	/// Delegate to be used to inform client of this UserControl that email addresses have been modified.
	/// </summary>
	public delegate void CustomerEmailAddressesModifiedDelegate(object sender, EventArgs eventArgs);

	/// <summary>
	/// Delegate to be used to inform client of this UserControl that the control has collapsed.
	/// </summary>
	public delegate void AdministrateCustomerEmailAddressesCollapsedDelegate(object sender, AdministrateCustomerEmailAddressesCollapsedEventArgs eventArgs);

	#endregion

	/// <summary>
	/// User control for administrating CC invoice email addresses.
	/// </summary>
	[ToolboxItem(true)]
	public partial class AdministrateCustomerEmailAddresses : UserControl
	{
		#region Private Delegates

		/// <summary>
		/// Delegate to be used for BeginInvoike.
		/// </summary>
		private delegate void InvokeDelegate();

		#endregion

		#region Constants

		/// <summary>
		/// Defines the key for the email number column.
		/// </summary>
		private const string _emailNrCell = "EmailNrString";

		/// <summary>
		/// Defines the key for the email address cell.
		/// </summary>
		private const string _emailaddressCell = "Email";

		/// <summary>
		/// Defines the template for displaying the email number.
		/// </summary>
		private const string _emailNrTemplate = "Email {0}";

		#endregion

		#region Events

		/// <summary>
		/// Occurs when control collapsed.
		/// </summary>
		public event AdministrateCustomerEmailAddressesCollapsedDelegate ControlCollapsed;

		/// <summary>
		/// Occurs when email addresses have been modified.
		/// </summary>
		public event CustomerEmailAddressesModifiedDelegate EmailAddressesModified;

		#endregion

		#region Member Variables

		private ControlState _currentControlState = new ControlState();

		/// <summary>
		/// Stores the order id.
		/// </summary>
		private int _orderId;

		/// <summary>
		/// Stores the target control.
		/// </summary>
		private Control _targetControl;

		/// <summary>
		/// Stores the message box manager.
		/// </summary>
		private readonly UltraMessageBoxManager _messageBoxManager = new UltraMessageBoxManager();

		/// <summary>
		/// Stores the bad cell i.e. the cell which has invalid email address.
		/// </summary>
		private UltraGridCell _badCell;

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets a value indicating whether this control is read only.
		/// </summary>
		/// <value><c>true</c> if control is read only; otherwise, <c>false</c>.</value>
		[Description("Defines if the user control is read only"), Category("Altova"), Browsable(false)]
		public bool ReadOnly { get; set; }

		/// <summary>
		/// Sets the target control i.e. the Control this User Control is targeting to.
		/// </summary>
		/// <value>The target control.</value>
		[Description("Sets the target control"), Category("Altova"), Browsable(false)]
		public Control TargetControl
		{
			set
			{
				_targetControl = value;

				// Subscrive for focus enter/leave events on target control
				_targetControl.Enter += OnTargetFocusEnter;
				_targetControl.Leave += OnTargetFocusLeave;
			}
		}

		/// <summary>
		/// Gets the email addresses.
		/// </summary>
		/// <value>The email addresses.</value>
		public BindingList<CustomerEmailAddress>? EmailAddresses
		{
			get { return emailAddressesBindingSource.DataSource as BindingList<CustomerEmailAddress>; }
		}

		private ControlState CurrentCurrentControlState
		{
			get { return _currentControlState; }
			set 
			{ 
				_currentControlState = value; 
                Trace.WriteLine($"Current control state changed to '{_currentControlState}'");
			}

		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor.
		/// </summary>
		public AdministrateCustomerEmailAddresses()
		{
			InitializeComponent();

			// Subscribe for opened and closed events on popup control container
			pccEmailAddresses.Opened += OnOpened;
			pccEmailAddresses.Closed += OnClosed;
		}

		#endregion

		#region Member Functions

		/// <summary>
		/// Shows the email addresses.
		/// </summary>
		/// <param name="orderId">The order id.</param>
		/// <param name="customerEmailAddresses">The customer email addresses.</param>
		public void ShowEmailAddresses(int orderId, BindingList<CustomerEmailAddress> customerEmailAddresses)
		{
			_orderId = orderId;

			// Set the list as data source for the grid
			emailAddressesBindingSource.DataSource = customerEmailAddresses;

			// Set this user control as popup control in popup control container.
			pccEmailAddresses.PopupControl = this;
		}

		#endregion

		#region Event Handler Functions

		/// <summary>
		/// Called when focus is entered on target control.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void OnTargetFocusEnter(object sender, EventArgs e)
		{
			Trace.WriteLine("OnTargetFocusEnter");
			if (CurrentCurrentControlState != ControlState.Closed)
				return;

			if (!ReadOnly)
			{
				// Get email address list from binding source
				var customerEmailAddresses = emailAddressesBindingSource.DataSource as BindingList<CustomerEmailAddress>;
				if (customerEmailAddresses != null)
				{
					// Add one placeholder order email address if needed
					if (customerEmailAddresses.Count == 0 || !customerEmailAddresses[customerEmailAddresses.Count - 1].IsEmpty)
					{
						customerEmailAddresses.Add(new CustomerEmailAddress(_orderId, string.Empty) {EmailNr = GetNextEmailNr()});
					}
				}

				// Set the consecutive email numbers
				SetEmailNumbers(customerEmailAddresses);

				Width = _targetControl.Width;

				// Get the screen position where the user control should be positioned relative to target control
				var screenPosition = _targetControl.PointToScreen(new Point(0, _targetControl.Size.Height + 2));
				try
				{
					CurrentCurrentControlState = ControlState.Opening;
					pccEmailAddresses.Show(screenPosition);
				}
				catch (Exception ex)
                {
                    Trace.WriteLine(ex.Message);
                    throw;
				}
			}
		}
		

		/// <summary>
		/// Called when focus leaves the target control.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void OnTargetFocusLeave(object sender, EventArgs e)
		{
			Trace.WriteLine("OnTargetFocusLeave");
			//Close();
		}

		/// <summary>
		/// Called when pupoup control container has been opened.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void OnOpened(object sender, EventArgs e)
		{
            Trace.WriteLine("OnOpened");
			CurrentCurrentControlState = ControlState.Opened;

			if (gvEmailAddresses.Rows.Count > 0)
			{
				var activeCell = gvEmailAddresses.Rows[gvEmailAddresses.Rows.Count - 1].Cells[_emailaddressCell];
				gvEmailAddresses.ActiveCell = activeCell;
			}
		}

		/// <summary>
		/// Called when popup control container has been closed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void OnClosed(object sender, EventArgs e)
		{
			_badCell = null;
			CurrentCurrentControlState = ControlState.Closed;

			// Get email addresses from binding source
			var customerEmailAddresses = emailAddressesBindingSource.DataSource as BindingList<CustomerEmailAddress>;
			if (customerEmailAddresses != null)
			{
				var emailAddressesToDelete = customerEmailAddresses.Where(customerEmailAddress => !customerEmailAddress.IsValid).ToList();

				// Remove all empty email addresses
				foreach (var emailAddress in emailAddressesToDelete)
				{
					customerEmailAddresses.Remove(emailAddress);
				}
			}

			// Do we have a client registered for Control Collapsed events?
			if (ControlCollapsed != null)
			{
				// Notify the client
				ControlCollapsed(this, new AdministrateCustomerEmailAddressesCollapsedEventArgs(_targetControl, customerEmailAddresses != null ? customerEmailAddresses.ToList() : null));
			}

			Trace.WriteLine("OnClosed");
		}

		/// <summary>
		/// Called when email address row is initialized.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Infragistics.Win.UltraWinGrid.InitializeRowEventArgs"/> instance containing the event data.</param>
		private void OnInitializeEmailAddressRow(object sender, InitializeRowEventArgs e)
		{
			// Get actual email address to be initialized
			var customerEmailAddress = e.Row.ListObject as CustomerEmailAddress;
			if (customerEmailAddress != null)
			{
				// Get email number cell
				var emailNrCell = e.Row.Cells[_emailNrCell];
				if (emailNrCell != null)
				{
					// Set appropriate email number in cell
					emailNrCell.Value = string.Format(_emailNrTemplate, customerEmailAddress.EmailNr);
				}
            }
		}

		/// <summary>
		/// Called when key down is detected in GridView.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyEventArgs"/> instance containing the event data.</param>
		private void OnKeyDown(object sender, KeyEventArgs e)
		{
			// React on Tab-Key but ignore it in combination with shift
			if (e.KeyCode == Keys.Tab && !e.Shift)
			{
				// Is the last row active?
				if (gvEmailAddresses.ActiveRow.Index == gvEmailAddresses.Rows.Count - 1)
				{
					// Is the cell with email address active?
					if (gvEmailAddresses.ActiveCell.Column.Key == _emailaddressCell)
					{
						// If the user didn't enter any email address in last cell it's an indication to close the popup
						if (string.IsNullOrEmpty(gvEmailAddresses.ActiveCell.Text))
						{
							Trace.WriteLine("No new Email address has been entered - closing will be started");
							Close();
						}
						else if (ValidateEmailAddress(gvEmailAddresses.ActiveCell.Text, gvEmailAddresses.ActiveCell))
						{
							// Valid email address => Add it to the list of email addresses
							emailAddressesBindingSource.Add(new CustomerEmailAddress(_orderId, string.Empty) {EmailNr = GetNextEmailNr()});
						}
						else
						{
							Trace.WriteLine("OnKeyDown: Email Address={gvEmailAddresses.ActiveCell.Text}");

							// Invalid email address => Stay in edit mode
							//gvEmailAddresses.ActiveCell.Value = string.Empty;
							gvEmailAddresses.PerformAction(UltraGridAction.EnterEditMode, false, false);
						}
					}
				}
			}
			else if (e.KeyCode == Keys.Delete)
			{
				if (gvEmailAddresses.ActiveRow != null)
				{
					int activeRowIndex = gvEmailAddresses.ActiveRow.Index;
					emailAddressesBindingSource.RemoveAt(gvEmailAddresses.ActiveRow.Index);
					if (activeRowIndex >= emailAddressesBindingSource.Count)
					{
						activeRowIndex--;
						if (activeRowIndex < 0)
							activeRowIndex = 0;
					}

					gvEmailAddresses.ActiveCell = gvEmailAddresses.Rows[activeRowIndex].Cells[_emailaddressCell];
				}
			}
		}

		/// <summary>
		/// Called when key up is detected in GridView.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.KeyEventArgs"/> instance containing the event data.</param>
		private void OnKeyUp(object sender, KeyEventArgs e)
		{
			// Has an invalid email address been entered?
			if (_badCell != null)
			{
				// Prevent further handling of the key
				e.Handled = true;

				// Set active cell to the cell with the invalid email address
				gvEmailAddresses.ActiveCell = _badCell;

				// Enter edit mode in GridView
				gvEmailAddresses.PerformAction(UltraGridAction.EnterEditMode, false, false);
				_badCell = null;
			}
		}

		/// <summary>
		/// Called when mouse up is detected in GridView.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
		private void OnMouseUp(object sender, MouseEventArgs e)
		{
			// Has an invalid email address been entered?
			if (_badCell != null)
			{
				// Set active cell to the cell with the invalid email address
				gvEmailAddresses.ActiveCell = _badCell;

				// Enter edit mode in GridView
				gvEmailAddresses.PerformAction(UltraGridAction.EnterEditMode, false, false);
				_badCell = null;
			}
		}

		/// <summary>
		/// Called when cell has been activated.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void OnCellActivated(object sender, EventArgs e)
		{
			Trace.WriteLine("OnCellActivated");
			Invoke(new InvokeDelegate(EnterEditMode));
		}

		/// <summary>
		/// Called when cell value has changed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Infragistics.Win.UltraWinGrid.CellEventArgs"/> instance containing the event data.</param>
		private void OnCellChange(object sender, CellEventArgs e)
		{
			// Get active cell from GridView
			var activeCell = gvEmailAddresses.ActiveCell;

			// Get actual text (part of email address)
			var actualText = activeCell.Text;
        }

		/// <summary>
		/// Called before cell is updated in GridView.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Infragistics.Win.UltraWinGrid.BeforeCellUpdateEventArgs"/> instance containing the event data.</param>
		private void OnBeforeCellUpdate(object sender, BeforeCellUpdateEventArgs e)
		{
			Trace.WriteLine(string.Format("OnBeforeCellUpdate: NewValue={0}", e.NewValue));

			// Is Email Address Cell updated?
			if (e.Cell.Column.Key == _emailaddressCell)
			{
				// Is it an invalid email address?
				if (_badCell != null || !ValidateEmailAddress(e.NewValue as string, e.Cell))
				{
					// Prevent further handling of this event
					e.Cancel = true;
				}
			}
		}

		/// <summary>
		/// Called when before cell is deactivated in GridView.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.ComponentModel.CancelEventArgs"/> instance containing the event data.</param>
		private void OnBeforeCellDeactivate(object sender, CancelEventArgs e)
		{
			// Has an invalid email address been entered?
			if (_badCell != null)
			{
				Trace.WriteLine("OnBeforeCellDeactivate, Bad Cell is present");

				// Prevent further handling of this event
				e.Cancel = true;

				var badCell = _badCell;
				_badCell = null;

				// Set active cell to the cell with the invalid email address
				gvEmailAddresses.ActiveCell = badCell;

				// Enter edit mode in GridView
				gvEmailAddresses.PerformAction(UltraGridAction.EnterEditMode, false, false);
			}
			else
			{
				Trace.WriteLine("OnBeforeCellDeactivate, Bad Cell is null");

				// Do we have a client registered for Email Address modification events?
				if (EmailAddressesModified != null)
				{
					// Notify client that there email addresses have been modified
					EmailAddressesModified(this, EventArgs.Empty);
				}
			}
		}

		/// <summary>
		/// Called to handle layout event.
		/// </summary>
		/// <remarks>When the popup control container is expanded for the first time GridView enters edit mode and leaves it immediately.
		/// This event is used as a trigger to enter edit mode again. This must be done asynchronously, otherwise it has no effect thus we
		/// do it using Invoke!</remarks>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.Windows.Forms.LayoutEventArgs"/> instance containing the event data.</param>
		private void OnLayout(object sender, LayoutEventArgs e)
		{
			if (e.AffectedProperty == "Visible")
			{
				Invoke(new InvokeDelegate(EnterEditMode));
			}
		}

		/// <summary>
		/// Called when user control dock has changed.
		/// </summary>
		/// <remarks>When the popup control container collapses and the last edited cell contains an invalid email address the message box
		/// showing the invalid email address would popup again. To prevent this this event is used to set flag isCollapsed which furthermore
		/// is checked before displaying the message box.</remarks>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		private void OnUserControlDockChanged(object sender, EventArgs e)
		{
			Trace.WriteLine("OnUserControlDockChanged");
			//if (CurrentCurrentControlState == ControlState.Opened)
			//	CurrentCurrentControlState = ControlState.Closing;
		}

		#endregion

		#region Helper Functions

		private void Close()
		{
			Trace.WriteLine("Function Close");
			if (CurrentCurrentControlState==ControlState.Opened)
			{
				CurrentCurrentControlState = ControlState.Closing;

				try
				{
					// Close the popup control container
					pccEmailAddresses.Close();
				}
				catch (Exception ex)
				{
					Trace.WriteLine("Exception when closing the popup control container");
				}
			}
			else
			{
				Trace.WriteLine(string.Format("Control is in state '{0}' - No further actions needed", CurrentCurrentControlState));
			}
		}

		/// <summary>
		/// Enters the edit mode in grid view.
		/// </summary>
		private void EnterEditMode()
		{
			gvEmailAddresses.PerformAction(UltraGridAction.EnterEditMode);
		}

		/// <summary>
		/// Validates the email address.
		/// </summary>
		/// <param name="emailAddress">The email address.</param>
		/// <param name="cell">The cell.</param>
		/// <returns>Returns true if email address is valid, otherwise false</returns>
		private bool ValidateEmailAddress(string emailAddress, UltraGridCell cell)
		{
			Trace.WriteLine("ValidateEmailAddress");
			if (!string.IsNullOrEmpty(emailAddress))
			{
				_messageBoxManager.ShowMessageBox(new UltraMessageBoxInfo(MessageBoxStyle.Default, null, $"{emailAddress} is not a valid email address", "Information", MessageBoxButtons.OK,
				                MessageBoxIcon.Information, MessageBoxDefaultButton.Button1, DefaultableBoolean.False));
				_badCell = cell;
				return false;
			}

			if (!IsUniqueEmailAddress(emailAddress))
			{
				_messageBoxManager.ShowMessageBox(new UltraMessageBoxInfo(MessageBoxStyle.Default, null, $"{emailAddress} is not a unique email address", "Information", MessageBoxButtons.OK,
				                MessageBoxIcon.Information, MessageBoxDefaultButton.Button1, DefaultableBoolean.False));
				_badCell = cell;
				return false;
			}

			return true;
		}

		private bool IsUniqueEmailAddress(string emailAddress)
		{
			var unique = true;
			var customerEmailAddresses = emailAddressesBindingSource.DataSource as BindingList<CustomerEmailAddress>;
			if (customerEmailAddresses != null)
			{
				unique = customerEmailAddresses.All(customerEmailAddress => string.Compare(emailAddress, customerEmailAddress.Email, StringComparison.OrdinalIgnoreCase) != 0);
			}
			return unique;
		}

		/// <summary>
		/// Gets the next email number.
		/// </summary>
		/// <returns>Returns the next email number</returns>
		private int GetNextEmailNr()
		{
			// Get email addresses from binding source
            if (emailAddressesBindingSource.DataSource is BindingList<CustomerEmailAddress> customerEmailAddresses)
			{
				var highestNr = 0;

				// Iterate through email address to get the highest number
				foreach (var customerEmailAddress in customerEmailAddresses)
				{
					if (customerEmailAddress.EmailNr.HasValue && customerEmailAddress.EmailNr.Value > highestNr)
					{
						highestNr = customerEmailAddress.EmailNr.Value;
					}
				}
				return ++highestNr;
			}

			return 0;
		}

		/// <summary>
		/// Sets the email numbers.
		/// </summary>
		/// <param name="customerEmailAddresses">The order email addresses.</param>
		private static void SetEmailNumbers(IEnumerable<CustomerEmailAddress> customerEmailAddresses)
		{
			if (customerEmailAddresses != null)
			{
				var nextNr = 1;

				// Iterate through email addresses and set consecutive email numbers
				foreach (var customerEmailAddress in customerEmailAddresses)
				{
					customerEmailAddress.EmailNr = nextNr++;
				}
			}
		}

        #endregion
    }

    /// <summary>
    /// Class AdministrateCustomerEmailAddressesCollapsedEventArgs comprises the arguments for the AdministrateCustomerEmailAddressesCollapsed event.
    /// </summary>
    public class AdministrateCustomerEmailAddressesCollapsedEventArgs : EventArgs
	{
		/// <summary>
		/// Gets or sets the target control.
		/// </summary>
		/// <value>The target control.</value>
		public Control TargetControl { get; private set; }

		/// <summary>
		/// Gets the customer email addresses.
		/// </summary>
		/// <value>
		/// The customer email addresses.
		/// </value>
		public List<CustomerEmailAddress> CustomerEmailAddresses { get; private set; }

		/// <summary>
		/// Initializes a new instance of the <see cref="AdministrateBillingEmailAddressesCollapsedEventArgs"/> class.
		/// </summary>
		/// <param name="targetControl">The target control.</param>
		/// <param name="customerEmailAddresses">The customer email addresses</param>
		public AdministrateCustomerEmailAddressesCollapsedEventArgs(Control targetControl, List<CustomerEmailAddress> customerEmailAddresses)
		{
			TargetControl = targetControl;
			CustomerEmailAddresses = customerEmailAddresses;
		}
	}
}
